# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import (
    List,
    Optional,
    Union,
)

import numpy as np

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import (
    ShapeT,
    shapeable,
    starts_with_zero,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    ArrayT,
    Checker,
    pipe,
    validated,
)

from .node_data import (
    Pwc,
    Stf,
    Target,
    Tensor,
)
from .utils import get_broadcasted_shape


def _validate_infidelity_input(
    hamiltonian: Pwc | Stf,
    target: Target,
    noise_operators: Optional[list[np.ndarray | Tensor | Pwc | Stf]],
) -> tuple[int, ...]:
    batch_shape = hamiltonian.batch_shape
    value_shape = hamiltonian.value_shape

    Checker.VALUE(
        value_shape == target.value_shape,
        "The Hamiltonian and the target must have the same value shape.",
        {
            "hamiltonian.value_shape": value_shape,
            "target.value_shape": target.value_shape,
        },
    )
    for idx, noise_operator in enumerate(noise_operators or []):
        if isinstance(noise_operator, (Pwc, Stf)):
            noise_operator_batch_shape = noise_operator.batch_shape
            noise_operator_value_shape = noise_operator.value_shape
        else:
            noise_operator_batch_shape = noise_operator.shape[:-2]
            noise_operator_value_shape = noise_operator.shape[-2:]

        Checker.VALUE(
            noise_operator_value_shape == value_shape,
            "The Hamiltonian and the noise operators must have the same value shape.",
            {
                "hamiltonian.value_shape": value_shape,
                f"noise_operators[{idx}] value shape": noise_operator_value_shape,
            },
        )

        batch_shape = get_broadcasted_shape(
            batch_shape,
            noise_operator_batch_shape,
            message="The batch shapes of the Hamiltonian and all noise operators must be "
            "broadcastable.",
        )

    return batch_shape


class InfidelityGraph:
    """
    Base class implementing infidelity graph methods.
    """

    @validated
    def target(
        self,
        operator: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
        filter_function_projector: Optional[
            Annotated[
                np.ndarray,
                pipe(ArrayT.NUMERIC().ndim(2).square().hermitian().idempotent()),
            ]
        ] = None,
    ) -> Target:
        r"""
        Create information about the target for system time evolution.

        Nodes created with this function contain two types of information: the
        target gate for the system time evolution, and the projection operator
        that defines the subspace of interest for robustness.

        Parameters
        ----------
        operator : np.ndarray or Tensor
            The target gate :math:`U_\mathrm{target}`. Must be a non-zero partial
            isometry.
        filter_function_projector : np.ndarray or None, optional
            The orthogonal projection matrix :math:`P` onto the subspace used for
            filter function calculations. If you provide a value then it must be
            Hermitian and idempotent. Defaults to the identity matrix.

        Returns
        -------
        Target
            The node containing the specified target information.

        See Also
        --------
        Graph.infidelity_pwc : Total infidelity of a system with a piecewise-constant Hamiltonian.
        Graph.infidelity_stf : Total infidelity of a system with a sampleable Hamiltonian.

        Notes
        -----
        The target gate :math:`U_\mathrm{target}` is a non-zero partial isometry,
        which means that it can be expressed in the form
        :math:`\sum_j \left|\psi_j\right>\left<\phi_j\right|`, where
        :math:`\left\{\left|\psi_j\right>\right\}` and
        :math:`\left\{\left|\phi_j\right>\right\}` both form (non-empty)
        orthonormal, but not necessarily complete, sets. Such a target represents
        a target state :math:`\left|\psi_j\right>` for each initial state
        :math:`\left|\phi_j\right>`. The resulting operational infidelity is 0
        if and only if, up to global phase, each initial state
        :math:`\left|\phi_j\right>` maps exactly to the corresponding final state
        :math:`\left|\psi_j\right>`.

        The filter function projector :math:`P` is an orthogonal projection
        matrix, which means that it satisfies :math:`P=P^\dagger=P^2`. The image
        of :math:`P` defines the set of initial states from which the calculated
        filter function measures robustness.

        Examples
        --------
        Define a target operation for the Pauli :math:`X` gate.

        >>> target_operation = graph.target(operator=np.array([[0, 1], [1, 0]]))
        >>> target_operation
        <Target: operation_name="target", value_shape=(2, 2)>

        See more examples in the `How to optimize controls robust to strong noise sources
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-robust-
        to-strong-noise-sources>`_ user guide.
        """

        shape = operator.shape

        if isinstance(operator, np.ndarray):
            Checker.VALUE(np.any(operator), "The operator must be non-zero.")
            Checker.VALUE(
                np.allclose(operator @ operator.T.conj() @ operator, operator),
                "The operator must be a partial isometry, but does not yield itself "
                "when multiplied by its adjoint and then itself.",
            )

        if filter_function_projector is not None:
            Checker.VALUE(
                filter_function_projector.shape == shape,
                "The projection operator must have the same dimension as the target operator.",
                {
                    "operator.shape": shape,
                    "filter_function_projector.shape": filter_function_projector.shape,
                },
            )

        operation = create_operation(self.target, locals())
        return Target(operation, value_shape=shape)

    @validated
    def infidelity_pwc(
        self,
        hamiltonian: Annotated[Pwc, pipe(after=ShapeT.OPERATOR())],
        target: Annotated[Target, pipe(after=ShapeT.OPERATOR())],
        noise_operators: Optional[
            List[
                Annotated[
                    Union[np.ndarray, Tensor, Pwc],
                    pipe(shapeable, after=ShapeT.OPERATOR()),
                ]
            ]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create the total infidelity of the given piecewise-constant system.

        Use this function to compute the sum of the operational infidelity (which
        measures how effectively the system achieves a target gate) and filter
        function values (which measure how robust the system evolution is to
        various perturbative noise processes). This total infidelity value
        provides a cost that measures how effectively and robustly a set of
        controls achieves a target operation.

        Note that the total infidelity returned by this function is at least zero,
        but might be larger than one (for example if the system is highly
        sensitive to one of the noise processes).

        Parameters
        ----------
        hamiltonian : Pwc
            The control Hamiltonian :math:`H_{\mathrm c}(t)`. You can provide
            either a single Hamiltonian or a batch of them.
        target : Target
            The object describing the target gate :math:`U_\mathrm{target}` and
            (optionally) the filter function projector :math:`P`. If you
            provide a batch of Hamiltonians, the function uses the same target
            for all the elements in the batch.
        noise_operators : list[np.ndarray or Tensor or Pwc] or None, optional
            The perturbative noise operators :math:`\{N_j(t)\}`. The operators
            in the list can either be single operators or batches of them. If
            any of the noise operators or the Hamiltonian are batches, the batch
            shapes must all be broadcastable. You can omit this list if there are
            no noises.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The total infidelity (operational infidelity plus filter function
            values) of the given system, with respect to the given target gate.
            If you provide a batch of Hamiltonians or noise operators, the
            function returns a batch of infidelities containing one infidelity
            for each Hamiltonian and list of noise operators in the input batches.

        Warnings
        --------
        The Hessian matrix cannot currently be calculated for a graph which includes
        an `infidelity_pwc` node if the `hamiltonian` has degenerate eigenvalues at
        any segment.

        See Also
        --------
        Graph.infidelity_stf : Corresponding operation for `Stf` Hamiltonians.
        Graph.target : Define the target operation of the time evolution.
        Graph.time_evolution_operators_pwc : Unitary time evolution operators
            for quantum systems with `Pwc` Hamiltonians.

        Notes
        -----
        The total system Hamiltonian is

        .. math:: H_{\mathrm c}(t) + \sum_j \beta_j(t) N_j(t),

        where :math:`\{\beta_j(t)\}` are small, dimensionless, stochastic
        variables.

        The total infidelity, as represented by this node, is the sum of the
        operational infidelity :math:`\mathcal{I}` and the filter functions
        :math:`\{F_j(0)\}` of each noise operator evaluated at zero frequency.

        The operational infidelity is

        .. math::
          \mathcal{I} = 1-\left|
            \frac{\mathrm{Tr} \left(U_\mathrm{target}^\dagger U(t)\right)}
            {\mathrm{Tr} \left(U_\mathrm{target}^\dagger U_\mathrm{target}\right)}
            \right|^2,

        where :math:`U(t)` is the unitary time evolution operator due to
        :math:`H_{\mathrm c}(t)`.

        The filter function for the noise operator :math:`N_j(t)` is a measure of
        robustness, defined at frequency :math:`f` as

        .. math::
          F_j(f) = \frac{1}{\mathrm{Tr}(P)} \mathrm{Tr} \left( P
            \mathcal{F} \left\{ \tilde N_j^\prime(t) \right\} \left[ \mathcal{F}
            \left\{ \tilde N^\prime (t) \right\} \right]^\dagger P \right),

        where :math:`\mathcal{F}` is the Fourier transform,
        :math:`\tilde N_j(t) \equiv U_c^\dagger(t) N_j(t) U_c(t)` is the
        toggling-frame noise operator, and
        :math:`\tilde N_j^\prime(t)\equiv
        \tilde N_j(t)-
        \frac{\mathrm{Tr}(P\tilde N_j(t)P)}{\mathrm{Tr}(P)} \mathbb{I}`
        differs from :math:`\tilde N_j(t)` only by a multiple of the identity but
        is trace-free on the subspace of interest. The filter function value at
        zero frequency quantifies the sensitivity of the controls to quasi-static
        noise applied via the corresponding noise operator.

        Examples
        --------
        Calculate infidelity of the identity gate for a noiseless single qubit.

        >>> sigma_z = np.array([[1, 0], [0, -1]])
        >>> hamiltonian = graph.pwc(
        ...     durations=np.array([0.1, 0.1]), values=np.array([sigma_z, -sigma_z])
        ... )
        >>> target = graph.target(np.eye(2))
        >>> infidelity = graph.infidelity_pwc(
        ...     hamiltonian=hamiltonian, target=target, name="infidelity"
        ... )
        >>> result = bo.execute_graph(graph=graph, output_node_names="infidelity")
        >>> result["output"]["infidelity"]["value"]
        0.0

        See more examples in the `How to optimize controls with non-linear dependencies
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-
        with-nonlinear-dependences>`_ user guide.
        """

        shape = _validate_infidelity_input(hamiltonian, target, noise_operators)

        operation = create_operation(self.infidelity_pwc, locals())
        return Tensor(operation, shape=shape)

    @validated
    def infidelity_stf(
        self,
        sample_times: Annotated[
            np.ndarray,
            pipe(ArrayT.REAL().ndim(1).ge(0).ascend(), after=starts_with_zero),
        ],
        hamiltonian: Annotated[Stf, pipe(after=ShapeT.OPERATOR())],
        target: Annotated[Target, pipe(after=ShapeT.OPERATOR())],
        noise_operators: Optional[
            List[
                Annotated[
                    Union[np.ndarray, Tensor, Stf],
                    pipe(shapeable, after=ShapeT.OPERATOR()),
                ]
            ]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create the total infidelity of a given system with a sampleable Hamiltonian.

        See :obj:`infidelity_pwc` for information about the total infidelity
        created by this function.

        Parameters
        ----------
        sample_times : np.ndarray(1D, real)
            The times at which the Hamiltonian and noise operators (if present) should be sampled
            for the integration. Must start with 0, be ordered, and contain at least one element.
        hamiltonian : Stf
            The control Hamiltonian :math:`H_{\mathrm c}(t)`. You can provide
            either a single Hamiltonian or a batch of them.
        target : Target
            The object describing the target gate :math:`U_\mathrm{target}` and
            (optionally) the filter function projector :math:`P`. If you
            provide a batch of Hamiltonians, the function uses the same target
            for all the elements in the batch.
        noise_operators : list[np.ndarray or Tensor or Stf] or None, optional
            The perturbative noise operators :math:`\{N_j(t)\}`. The operators
            in the list can either be single operators or batches of them. If
            any of the noise operators or the Hamiltonian are batches, the batch
            shapes must all be broadcastable. You can omit this list if there are
            no noises.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The total infidelity (operational infidelity plus filter function
            values) of the given system, with respect to the given target gate,
            at the last time in `sample_times`.
            If you provide a batch of Hamiltonians or noise operators, the
            function returns a batch of infidelities containing one infidelity
            for each Hamiltonian and list of noise operators in the input batches.

        See Also
        --------
        Graph.infidelity_pwc : Corresponding operation for `Pwc` Hamiltonians.
        Graph.target : Define the target operation of the time evolution.
        Graph.time_evolution_operators_stf : Unitary time evolution operators
            for quantum systems with `Stf` Hamiltonians.

        Examples
        --------
        Calculate the infidelity of the Pauli :math:`X` gate for a noiseless qubit.

        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> hamiltonian = graph.constant_stf_operator(np.pi * sigma_x / 2)
        >>> target = graph.target(sigma_x)
        >>> infidelity = graph.infidelity_stf(
        ...     sample_times=np.linspace(0, 0.5, 100),
        ...     hamiltonian=hamiltonian,
        ...     target=target,
        ...     name="infidelity",
        ... )
        >>> result = bo.execute_graph(graph=graph, output_node_names="infidelity")
        >>> result["output"]["infidelity"]["value"]
        0.5000000000260991
        """
        shape = _validate_infidelity_input(hamiltonian, target, noise_operators)

        operation = create_operation(self.infidelity_stf, locals())
        return Tensor(operation, shape=shape)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
STF signal library nodes.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
)

import numpy as np

from boulderopal._nodes.node_data import (
    Stf,
    Tensor,
)
from boulderopal._nodes.validation import (
    ShapeT,
    positive_scalar,
    scalar,
    shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    Checker,
    ScalarT,
    pipe,
    validated,
)

if TYPE_CHECKING:
    from boulderopal.graph._graph import Graph


class StfSignals:
    """
    Base class implementing Stf signal graph methods.
    """

    def __init__(self, graph: Graph) -> None:
        self._graph = graph

    @validated
    def sech_pulse_stf(
        self,
        amplitude: Annotated[Union[float, complex, Tensor], pipe(scalar)],
        width: Annotated[Union[float, Tensor], pipe(positive_scalar)],
        center_time: Annotated[Union[float, Tensor], pipe(scalar)],
    ) -> Stf:
        r"""
        Create an `Stf` representing a hyperbolic secant pulse.

        Parameters
        ----------
        amplitude : float or complex or Tensor
            The amplitude of the pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        width : float or Tensor
            The characteristic time for the hyperbolic secant pulse, :math:`t_\mathrm{pulse}`.
            It must either be a scalar or contain a single element.
        center_time : float or Tensor
            The time at which the pulse peaks, :math:`t_\mathrm{peak}`.
            It must either be a scalar or contain a single element.

        Returns
        -------
        Stf
            The sampleable hyperbolic secant pulse.

        See Also
        --------
        :func:`Graph.signals.gaussian_pulse_stf <signals.gaussian_pulse_stf>`
            Create an `Stf` representing a Gaussian pulse.
        :func:`boulderopal.signals.sech_pulse`
            Create a `Signal` object representing a hyperbolic secant pulse.
        :func:`Graph.signals.sech_pulse_pwc <signals.sech_pulse_pwc>`
            Corresponding operation with `Pwc` output.

        Notes
        -----
        The hyperbolic secant pulse is defined as

            .. math:: \mathop{\mathrm{Sech}}(t)
                = \frac{A}{\cosh\left((t - t_\mathrm{peak}) / t_\mathrm{pulse} \right)} .

        The full width at half maximum of the pulse is about :math:`2.634 t_\mathrm{pulse}`.

        Examples
        --------
        Define a sampleable sech pulse.

        >>> sech = graph.signals.sech_pulse_stf(
        ...     amplitude=1.0, width=0.1, center_time=0.5
        ... )
        >>> graph.discretize_stf(stf=sech, duration=1.2, segment_count=5, name="sech")
        <Pwc: name="sech", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sech")
        >>> result["output"]["sech"]
        {'durations': array([0.24, 0.24, 0.24, 0.24, 0.24]),
        'values': array([0.04471916, 0.46492199, 0.64805427, 0.06667228, 0.00605505]),
        'time_dimension': 0}

        Define a sampleable sech pulse with optimizable parameters.

        >>> amplitude = graph.optimizable_scalar(
        ...     lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> width = graph.optimizable_scalar(
        ...     lower_bound=0.1, upper_bound=0.5, name="width"
        ... )
        >>> center_time = graph.optimizable_scalar(
        ...     lower_bound=0.2, upper_bound=0.8, name="center_time"
        ... )
        >>> graph.signals.sech_pulse_stf(
        ...     amplitude=amplitude, width=width, center_time=center_time
        ... )
        <Stf: operation_name="truediv", value_shape=(), batch_shape=()>
        """

        return amplitude / self._graph.cosh((self._graph.identity_stf() - center_time) / width)

    @validated
    def gaussian_pulse_stf(
        self,
        amplitude: Annotated[Union[float, complex, Tensor], pipe(scalar)],
        width: Annotated[Union[float, Tensor], pipe(positive_scalar)],
        center_time: Annotated[Union[float, Tensor], pipe(scalar)],
        drag: Optional[Annotated[Union[float, Tensor], pipe(scalar)]] = None,
    ) -> Stf:
        r"""
        Create an `Stf` representing a Gaussian pulse.

        Parameters
        ----------
        amplitude : float or complex or Tensor
            The amplitude of the Gaussian pulse, :math:`A`.
            It must either be a scalar or contain a single element.
        width : float or Tensor
            The standard deviation of the Gaussian pulse, :math:`\sigma`.
            It must either be a scalar or contain a single element.
        center_time : float or Tensor
            The center of the Gaussian pulse, :math:`t_0`.
            It must either be a scalar or contain a single element.
        drag : float or Tensor or None, optional
            The DRAG parameter, :math:`\beta`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to no DRAG correction.

        Returns
        -------
        Stf
            The sampleable Gaussian pulse.

        See Also
        --------
        :func:`Graph.signals.gaussian_pulse_pwc <signals.gaussian_pulse_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.signals.sech_pulse_stf <signals.sech_pulse_stf>`
            Create an `Stf` representing a hyperbolic secant pulse.

        Notes
        -----
        The Gaussian pulse is defined as

        .. math:: \mathop{\mathrm{Gaussian}}(t) =
            A \left(1-\frac{i\beta (t-t_0)}{\sigma^2}\right)
            \exp \left(- \frac{(t-t_0)^2}{2\sigma^2} \right) .

        Examples
        --------
        Define a sampleable Gaussian pulse.

        >>> gaussian = graph.signals.gaussian_pulse_stf(
        ...     amplitude=1.0, width=0.1, center_time=0.5
        ... )
        >>> graph.discretize_stf(
        ...     gaussian, duration=1, segment_count=5, name="gaussian"
        ... )
        <Pwc: name="gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="gaussian")
        >>> result["output"]["gaussian"]
        {'durations': array([0.2, 0.2, 0.2, 0.2, 0.2]),
        'values': array([3.35462628e-04, 1.35335283e-01, 1.00000000e+00, 1.35335283e-01,
                3.35462628e-04]),
        'time_dimension': 0}

        Define a sampleable Gaussian with a DRAG correction.

        >>> drag_gaussian = graph.signals.gaussian_pulse_stf(
        ...     amplitude=1.0, width=0.1, center_time=0.5, drag=0.2
        ... )
        >>> graph.discretize_stf(
        ...     drag_gaussian, duration=1, segment_count=5, name="drag_gaussian"
        ... )
        <Pwc: name="drag_gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="drag_gaussian")
        >>> result["output"]["drag_gaussian"]
        {'durations': array([0.2, 0.2, 0.2, 0.2, 0.2]),
        'values': array([3.35462628e-04+0.0026837j , 1.35335283e-01+0.54134113j,
                1.00000000e+00+0.j        , 1.35335283e-01-0.54134113j,
                3.35462628e-04-0.0026837j ]),
        'time_dimension': 0}

        Define a sampleable Gaussian pulse with optimizable parameters.

        >>> amplitude = graph.optimizable_scalar(
        ...     lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
        ... )
        >>> width = graph.optimizable_scalar(
        ...     lower_bound=0.1, upper_bound=0.5, name="width"
        ... )
        >>> center_time = graph.optimizable_scalar(
        ...     lower_bound=0.2, upper_bound=0.8, name="center_time"
        ... )
        >>> drag = graph.optimizable_scalar(
        ...     lower_bound=0, upper_bound=0.5, name="drag"
        ... )
        >>> graph.signals.gaussian_pulse_stf(
        ...     amplitude=amplitude, width=width, center_time=center_time, drag=drag
        ... )
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        """
        if drag is not None:
            correction = -(1j * drag / (width**2)) * (self._graph.identity_stf() - center_time)
            amplitude *= 1.0 + correction

        return amplitude * self._graph.exp(
            -((self._graph.identity_stf() - center_time) ** 2) / (2 * width**2),
        )

    @validated
    def sinusoid_stf(
        self,
        amplitude: Annotated[Union[float, complex, Tensor], pipe(scalar)],
        angular_frequency: Annotated[Union[float, Tensor], pipe(positive_scalar)],
        phase: Annotated[Union[float, Tensor], pipe(scalar)] = 0.0,
    ) -> Stf:
        r"""
        Create an `Stf` representing a sinusoidal oscillation.

        Parameters
        ----------
        amplitude : float or complex or Tensor
            The amplitude of the oscillation, :math:`A`.
            It must either be a scalar or contain a single element.
        angular_frequency : float or Tensor
            The angular frequency of the oscillation, :math:`\omega`.
            It must either be a scalar or contain a single element.
        phase : float or Tensor, optional
            The phase of the oscillation, :math:`\phi`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to 0.

        Returns
        -------
        Stf
            The sampleable sinusoid.

        See Also
        --------
        :func:`Graph.signals.hann_series_stf <signals.hann_series_stf>`
            Create an `Stf` representing a sum of Hann window functions.
        :func:`boulderopal.signals.sinusoid`
            Create a `Signal` object representing a sinusoidal oscillation.
        :func:`Graph.signals.sinusoid_pwc <signals.sinusoid_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.sin`
            Calculate the element-wise sine of an object.

        Notes
        -----
        The sinusoid is defined as

        .. math:: \mathop{\mathrm{Sinusoid}}(t) = A \sin \left( \omega t + \phi \right) .

        Examples
        --------
        Define an STF oscillation.

        >>> oscillation = graph.signals.sinusoid_stf(
        ...     amplitude=2.0, angular_frequency=3.0, phase=np.pi/4
        ... )
        >>> graph.discretize_stf(
        ...     oscillation, duration=10, segment_count=5, name="oscillation"
        ... )
        <Pwc: name="oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="oscillation")
        >>> result["output"]["oscillation"]
        {'durations': array([2., 2., 2., 2., 2.]),
        'values': array([-1.20048699, -0.70570922, -0.15471507,  0.4086036 ,  0.93937314]),
        'time_dimension': 0}

        Define a sinusoid with optimizable parameters.

        >>> amplitude = graph.optimizable_scalar(
        ...     lower_bound=0, upper_bound=4e3, name="amplitude"
        ... )
        >>> angular_frequency = graph.optimizable_scalar(
        ...     lower_bound=5e6, upper_bound=20e6, name="angular_frequency"
        ... )
        >>> phase = graph.optimization_variable(
        ...     count=1,
        ...     lower_bound=0,
        ...     upper_bound=2*np.pi,
        ...     is_lower_unbounded=True,
        ...     is_upper_unbounded=True,
        ...     name="phase",
        ... )
        >>> graph.signals.sinusoid_stf(
        ...     amplitude=amplitude, angular_frequency=angular_frequency, phase=phase
        ... )
        <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
        """
        return amplitude * self._graph.sin(angular_frequency * self._graph.identity_stf() + phase)

    @validated
    def hann_series_stf(
        self,
        coefficients: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.VECTOR().no_batch()),
        ],
        end_time: Annotated[float, pipe(ScalarT.REAL())],
        start_time: Annotated[float, pipe(ScalarT.REAL())] = 0.0,
    ) -> Stf:
        r"""
        Create an `Stf` representing a sum of Hann window functions.

        Parameters
        ----------
        coefficients : np.ndarray or Tensor
            The coefficients for the different Hann window functions, :math:`c_n`.
            It must be a 1D array or Tensor.
        end_time : float
            The time at which the Hann series ends, :math:`t_\mathrm{end}`.
        start_time : float, optional
            The time at which the Hann series starts, :math:`t_\mathrm{start}`.
            Defaults to 0.

        Returns
        -------
        Stf
            The sampleable Hann window functions series.

        See Also
        --------
        :func:`boulderopal.signals.hann_series`
            Create a `Signal` object representing a sum of Hann window functions.
        :func:`Graph.signals.hann_series_pwc <signals.hann_series_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.signals.sinusoid_stf <signals.sinusoid_stf>`
            Create an `Stf` representing a sinusoidal oscillation.

        Notes
        -----
        The series is defined as

        .. math:: \mathop{\mathrm{Hann}}(t)
            = \sum_{n=1}^N c_n \sin^2 \left(
                \frac{\pi n (t - t_\mathrm{start})}{t_\mathrm{end} - t_\mathrm{start}}
            \right) ,

        where :math:`N` is the number of coefficients.

        Note that the function values outside the :math:`(t_\mathrm{start}, t_\mathrm{end})` range
        will not be zero.

        Examples
        --------
        Define a simple sampleable Hann series.

        >>> hann = graph.signals.hann_series_stf(
        ...     coefficients=np.array([0.5, 1, 0.25]), end_time=1.0
        ... )
        >>> graph.discretize_stf(hann, duration=1, segment_count=5, name="hann")
        <Pwc: name="hann", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="hann")
        >>> result["output"]["hann"]
        {'durations': array([0.2, 0.2, 0.2, 0.2, 0.2]),
        'values': array([0.5568644 , 1.25563559, 0.75      , 1.25563569, 0.55686415]),
        'time_dimension': 0}

        Define a sampleable Hann series with optimizable coefficients.

        >>> coefficients = graph.optimization_variable(
        ...     count=8, lower_bound=-3.5e6, upper_bound=3.5e6, name="coefficients"
        ... )
        >>> graph.signals.hann_series_stf(coefficients=coefficients, end_time=2.0e-6)
        <Stf: operation_name="stf_sum", value_shape=(), batch_shape=()>
        """

        Checker.VALUE(
            end_time > start_time,
            "The end time must be greater than the start time.",
            {"start_time": start_time, "end_time": end_time},
        )

        # Define scaled times π (t - t_start) / (t_end - t_start).
        scaled_time = (self._graph.identity_stf() - start_time) * (np.pi / (end_time - start_time))

        # Calculate function values.
        stfs = [
            coefficients[idx] * self._graph.sin((idx + 1) * scaled_time) ** 2
            for idx in range(coefficients.shape[0])
        ]
        return self._graph.stf_sum(stfs)

    @validated
    def linear_ramp_stf(
        self,
        slope: Annotated[Union[float, complex, Tensor], pipe(scalar)],
        shift: Annotated[Union[float, complex, Tensor], pipe(scalar)] = 0.0,
    ) -> Stf:
        r"""
        Create an `Stf` representing a linear ramp.

        Parameters
        ----------
        slope : float or complex or Tensor
            The slope of the ramp, :math:`a`.
            It must either be a scalar or contain a single element.
        shift : float or complex or Tensor, optional
            The value of the ramp at :math:`t = 0`, :math:`b`.
            It must either be a scalar or contain a single element.
            Defaults to 0.

        Returns
        -------
        Stf
            The sampleable linear ramp.

        See Also
        --------
        :func:`boulderopal.signals.linear_ramp`
            Create a `Signal` object representing a linear ramp.
        :func:`Graph.signals.linear_ramp_pwc <signals.linear_ramp_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.signals.tanh_ramp_stf <signals.tanh_ramp_stf>`
            Create an `Stf` representing a hyperbolic tangent ramp.

        Notes
        -----
        The linear ramp is defined as

        .. math:: \mathop{\mathrm{Linear}}(t) = a t + b .

        Examples
        --------
        Define a linear STF ramp.

        >>> linear = graph.signals.linear_ramp_stf(slope=4.0, shift=-2.0)
        >>> graph.discretize_stf(linear, duration=1, segment_count=5, name="linear")
        <Pwc: name="linear", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="linear")
        >>> result["output"]["linear"]
        {'durations': array([0.2, 0.2, 0.2, 0.2, 0.2]),
        'values': array([-1.6, -0.8,  0. ,  0.8,  1.6]),
        'time_dimension': 0}

        Define a linear STF ramp with an optimizable slope and root.

        >>> slope = graph.optimizable_scalar(
        ...     lower_bound=-4, upper_bound=4, name="slope"
        ... )
        >>> root = graph.optimizable_scalar(
        ...     lower_bound=-4, upper_bound=4, name="slope"
        ... )
        >>> shift = - slope * root
        >>> graph.signals.linear_ramp_stf(slope=slope, shift=shift)
        <Stf: operation_name="add", value_shape=(), batch_shape=()>
        """
        return slope * self._graph.identity_stf() + shift

    @validated
    def tanh_ramp_stf(
        self,
        center_time: Annotated[Union[float, Tensor], pipe(scalar)],
        ramp_duration: Annotated[Union[float, Tensor], pipe(positive_scalar)],
        end_value: Annotated[Union[float, complex, Tensor], pipe(scalar)],
        start_value: Optional[Annotated[Union[float, complex, Tensor], pipe(scalar)]] = None,
    ) -> Stf:
        r"""
        Create an `Stf` representing a hyperbolic tangent ramp.

        Parameters
        ----------
        center_time : float or Tensor
            The time at which the ramp has its greatest slope, :math:`t_0`.
            It must either be a scalar or contain a single element.
        ramp_duration : float or Tensor
            The characteristic time for the hyperbolic tangent ramp, :math:`t_\mathrm{ramp}`.
            It must either be a scalar or contain a single element.
        end_value : float or complex or Tensor
            The asymptotic value of the ramp towards :math:`t \to +\infty`, :math:`a_+`.
            It must either be a scalar or contain a single element.
        start_value : float or complex or Tensor, optional
            The asymptotic value of the ramp towards :math:`t \to -\infty`, :math:`a_-`.
            If passed, it must either be a scalar or contain a single element.
            Defaults to minus `end_value`.

        Returns
        -------
        Stf
            The sampleable hyperbolic tangent ramp.

        See Also
        --------
        :func:`Graph.signals.linear_ramp_stf <signals.linear_ramp_stf>`
            Create an `Stf` representing a linear ramp.
        :func:`boulderopal.signals.tanh_ramp`
            Create a `Signal` object representing a hyperbolic tangent ramp.
        :func:`Graph.signals.tanh_ramp_pwc <signals.tanh_ramp_pwc>`
            Corresponding operation with `Pwc` output.
        :func:`Graph.tanh`
            Calculate the element-wise hyperbolic tangent of an object.

        Notes
        -----
        The hyperbolic tangent ramp is defined as

        .. math:: \mathop{\mathrm{Tanh}}(t)
            = \frac{a_+ + a_-}{2}
                + \frac{a_+ - a_-}{2} \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

        where the function's asymptotic values :math:`a_\pm` are defined by:

        .. math::  a_\pm := \lim_{t\to\pm\infty} \mathop{\mathrm{Tanh}}(t) ,

        and :math:`t_0` is related to :math:`t_\mathrm{ramp}` by:

        .. math::
            \left.\frac{{\mathrm d}\mathop{\mathrm{Tanh}}(t)}{{\mathrm d}t}\right|_{t=t_0}
                = \frac{ (a_+ - a_-)}{2 t_\mathrm{ramp}} .

        With the default value of `start_value` (:math:`a_-`),
        the ramp expression simplifies to

        .. math:: \mathop{\mathrm{Tanh}}(t)
            = A \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

        where :math:`A = a_+` is the end value (the start value is then :math:`-A`).

        Examples
        --------
        Define a simple sampleable hyperbolic tangent ramp.

        >>> tanh = graph.signals.tanh_ramp_stf(
        ...     center_time=0.4, ramp_duration=0.2, end_value=2, start_value=-1
        ... )
        >>> graph.discretize_stf(tanh, duration=1, segment_count=5, name="tanh")
        <Pwc: name="tanh", operation_name="discretize_stf", value_shape=(), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="tanh")
        >>> result["output"]["tanh"]
        {'durations': array([0.2, 0.2, 0.2, 0.2, 0.2]),
        'values': array([-0.85772238, -0.19317574,  1.19317574,  1.85772238,  1.97992145]),
        'time_dimension': 0}

        Define a hyperbolic tangent ramp with optimizable parameters.

        >>> center_time = graph.optimizable_scalar(
        ...     lower_bound=0.25e-6, upper_bound=0.75e-6, name="center_time"
        ... )
        >>> ramp_duration = graph.optimizable_scalar(
        ...     lower_bound=0.1e-6, upper_bound=0.3e-6, name="ramp_duration"
        ... )
        >>> end_value = graph.optimizable_scalar(
        ...     lower_bound=0, upper_bound=3e6, name="end_value"
        ... )
        >>> graph.signals.tanh_ramp_stf(
        ...     center_time=center_time, ramp_duration=ramp_duration, end_value=end_value
        ... )
        <Stf: operation_name="add", value_shape=(), batch_shape=()>
        """

        if start_value is None:
            start_value = -end_value

        return start_value + 0.5 * (end_value - start_value) * (
            1 + self._graph.tanh((self._graph.identity_stf() - center_time) / ramp_duration)
        )



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import (
    List,
    Optional,
    Union,
)

import numpy as np
from pydantic import Field
from scipy.sparse import (
    issparse,
    spmatrix,
)

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import (
    ArrayT,
    ShapeT,
    _to_scalar_tensor,
    bounded_by,
    shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    Checker,
    ScalarT,
    pipe,
    type_pipe,
    validated,
)

from .node_data import (
    Pwc,
    SparsePwc,
    Tensor,
)
from .utils import mesh_pwc_durations


def _int_messenger(name: str) -> str:
    return f"The {name} must be a positive int or a scalar Tensor."


def _float_messenger(name: str) -> str:
    return f"The {name} must be a positive float or a scalar Tensor."


_scalar_int = type_pipe([ScalarT.INT().gt(0), _to_scalar_tensor], _int_messenger)
_scalar_float = type_pipe([ScalarT.REAL().gt(0), _to_scalar_tensor], _float_messenger)


def _validate_terms(value: list[SparsePwc], *, name: str) -> list[SparsePwc]:
    """
    Validate a list of SparsePwc terms.
    """
    shape_0 = value[0].value_shape
    for i, term in enumerate(value):
        shape = term.value_shape
        Checker.VALUE(
            shape == shape_0,
            f"All the elements in {name} must have the same shape.",
            {"terms[0].value_shape": shape_0, f"terms[{i}].value_shape": shape},
        )
    return value


class SparseGraph:
    """
    Base class implementing sparse graph methods.
    """

    @validated
    def sparse_pwc_operator(
        self,
        signal: Annotated[Pwc, pipe(shapeable, after=ShapeT.SIGNAL().no_batch())],
        operator: Annotated[
            Union[np.ndarray, spmatrix, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
    ) -> SparsePwc:
        r"""
        Create a sparse piecewise-constant operator (sparse-matrix-valued function of time).

        Each of the piecewise-constant segments (time periods) is a scalar multiple
        of the operator.

        Parameters
        ----------
        signal : Pwc
            The scalar-valued piecewise-constant function of time :math:`a(t)`.
        operator : numpy.ndarray or scipy.sparse.spmatrix or Tensor
            The sparse operator :math:`A` to be scaled over time.
            If you pass a Tensor or NumPy array, it will be internally
            converted into a sparse representation.

        Returns
        -------
        SparsePwc
            The piecewise-constant sparse operator :math:`a(t)A`.

        See Also
        --------
        Graph.constant_sparse_pwc_operator : Create constant `SparsePwc`\s.
        Graph.density_matrix_evolution_pwc : Evolve a quantum state in an open system.
        Graph.pwc_operator : Corresponding operation for `Pwc`\s.
        Graph.sparse_pwc_hermitian_part : Hermitian part of a `SparsePwc` operator.
        Graph.sparse_pwc_sum : Sum multiple `SparsePwc`\s.
        Graph.state_evolution_pwc : Evolve a quantum state.

        Examples
        --------
        Create a sparse PWC operator.

        >>> from scipy.sparse import coo_matrix
        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> signal = graph.pwc_signal(values=np.array([1, 2, 3]), duration=0.1)
        >>> graph.sparse_pwc_operator(signal=signal, operator=coo_matrix(sigma_x))
        <SparsePwc: operation_name="sparse_pwc_operator", value_shape=(2, 2)>

        See more examples in the `How to simulate large open system dynamics
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
        user guide.
        """
        operation = create_operation(self.sparse_pwc_operator, locals())
        return SparsePwc(operation, value_shape=operator.shape, durations=signal.durations)

    @validated
    def constant_sparse_pwc_operator(
        self,
        duration: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        operator: Annotated[
            Union[np.ndarray, spmatrix, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
    ) -> SparsePwc:
        r"""
        Create a constant sparse piecewise-constant operator over a specified duration.

        Parameters
        ----------
        duration : float
            The duration :math:`\tau` for the resulting piecewise-constant operator.
        operator : numpy.ndarray or scipy.sparse.spmatrix or Tensor
            The sparse operator :math:`A`. If you pass a Tensor or NumPy array,
            it will be internally converted into a sparse representation.

        Returns
        -------
        SparsePwc
            The constant operator :math:`t\mapsto A` (for :math:`0\leq t\leq\tau`).

        See Also
        --------
        Graph.constant_pwc_operator : Corresponding operation for `Pwc`\s.
        Graph.sparse_pwc_hermitian_part : Hermitian part of a `SparsePwc` operator.
        Graph.sparse_pwc_operator : Create `SparsePwc` operators.
        Graph.sparse_pwc_sum : Sum multiple `SparsePwc`\s.

        Examples
        --------
        Create a constant sparse PWC operator.

        >>> from scipy.sparse import coo_matrix
        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_x))
        <SparsePwc: operation_name="constant_sparse_pwc_operator", value_shape=(2, 2)>

        See more examples in the `How to simulate large open system dynamics
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
        user guide.
        """
        operation = create_operation(self.constant_sparse_pwc_operator, locals())
        return SparsePwc(operation, value_shape=operator.shape, durations=np.array([duration]))

    @validated
    def sparse_pwc_sum(
        self,
        terms: Annotated[
            List[Annotated[SparsePwc, pipe(shapeable, after=ShapeT.OPERATOR().no_batch())]],
            Field(min_length=1),
            pipe(after=_validate_terms),
        ],
    ) -> SparsePwc:
        r"""
        Create the sum of multiple sparse-matrix-valued piecewise-constant functions.

        Parameters
        ----------
        terms : list[SparsePwc]
            The individual piecewise-constant terms :math:`\{v_j(t)\}` to sum.
            All terms must be sparse, have values of the same shape,
            and have the same total duration
            but may have different numbers of segments of different durations.

        Returns
        -------
        SparsePwc
            The piecewise-constant function of time :math:`\sum_j v_j(t)`. It
            has the same shape as each of the `terms` that you provided.

        See Also
        --------
        Graph.constant_sparse_pwc_operator : Create constant `SparsePwc`\s.
        Graph.pwc_sum : Corresponding operation for `Pwc`\s.
        Graph.sparse_pwc_operator : Create `SparsePwc` operators.

        Examples
        --------
        Sum two sparse PWC operators.

        >>> from scipy.sparse import coo_matrix
        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> sigma_y = np.array([[0, -1j], [1j, 0]])
        >>> sp_x = graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_x))
        >>> sp_y = graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_y))
        >>> graph.sparse_pwc_sum([sp_x, sp_y])
        <SparsePwc: operation_name="sparse_pwc_sum", value_shape=(2, 2)>

        See more examples in the `How to simulate large open system dynamics
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-system-dynamics>`_
        user guide.
        """
        value_shape = terms[0].value_shape
        operation = create_operation(self.sparse_pwc_sum, locals())
        return SparsePwc(operation, value_shape=value_shape, durations=mesh_pwc_durations(terms))

    @validated
    def sparse_pwc_hermitian_part(
        self,
        operator: Annotated[SparsePwc, pipe(shapeable, after=ShapeT.OPERATOR().no_batch())],
    ) -> SparsePwc:
        r"""
        Create the Hermitian part of a piecewise-constant operator.

        Parameters
        ----------
        operator : SparsePwc
            The operator :math:`A(t)`.

        Returns
        -------
        SparsePwc
            The Hermitian part :math:`\frac{1}{2}(A(t)+A^\dagger(t))`.

        See Also
        --------
        Graph.hermitian_part : Hermitian part of an operator.
        Graph.sparse_pwc_operator : Create `SparsePwc`\s.

        Examples
        --------
        Create a Hermitian sparse PWC operator.

        >>> from scipy.sparse import coo_matrix
        >>> sigma_m = np.array([[0, 1], [0, 0]])
        >>> sp_m = graph.constant_sparse_pwc_operator(duration=0.1, operator=coo_matrix(sigma_m))
        >>> graph.sparse_pwc_hermitian_part(sp_m)
        <SparsePwc: operation_name="sparse_pwc_hermitian_part", value_shape=(2, 2)>
        """
        operation = create_operation(self.sparse_pwc_hermitian_part, locals())
        return SparsePwc(operation, value_shape=operator.value_shape, durations=operator.durations)

    @validated
    def state_evolution_pwc(
        self,
        initial_state: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.VECTOR().no_batch()),
        ],
        hamiltonian: Annotated[
            Union[Pwc, SparsePwc],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
        krylov_subspace_dimension: Annotated[Union[int, Tensor], pipe(_scalar_int)],
        sample_times: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the time evolution of a state generated by a piecewise-constant
        Hamiltonian by using the Lanczos method.

        Parameters
        ----------
        initial_state : Tensor or np.ndarray
            The initial state as a Tensor or np.ndarray of shape ``(D,)``.
        hamiltonian : Pwc or SparsePwc
            The control Hamiltonian. Uses sparse matrix multiplication if of type
            `SparsePwc`, which can be more efficient for large operators that are
            relatively sparse (contain mostly zeros).
        krylov_subspace_dimension : Tensor or int
            The dimension of the Krylov subspace `k` for the Lanczos method.
        sample_times : np.ndarray(1D, real) or None, optional
            The N times at which you want to sample the state. Must be ordered and
            contain at least one element. If omitted only the evolved state at the
            final time of the control Hamiltonian is returned.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            Tensor of shape ``(N, D)`` or ``(D,)`` if `sample_times` is omitted
            representing the state evolution. The n-th element (along the first
            dimension) represents the state at ``sample_times[n]`` evolved from
            the initial state.

        Warnings
        --------
        This calculation can be relatively inefficient for small systems (very roughly
        speaking, when the dimension of your Hilbert space is less than around 100; the
        exact cutoff depends on the specifics of your problem though). You should generally
        first try using `time_evolution_operators_pwc` to get the full time evolution
        operators (and evolve your state using those), and only switch to this method
        if that approach proves too slow or memory intensive. See the
        `How to simulate quantum dynamics for noiseless systems using graphs`_ user guide
        for an example of calculating state evolution with `time_evolution_operators_pwc`.

        .. _How to simulate quantum dynamics for noiseless systems using graphs:
            https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-quantum-
            dynamics-for-noiseless-systems-using-graphs

        See Also
        --------
        Graph.density_matrix_evolution_pwc : Corresponding operation for open systems.
        Graph.discretize_stf : Discretize an `Stf` into a `Pwc`.
        Graph.estimated_krylov_subspace_dimension_lanczos :
            Obtain a Krylov subspace dimension to use with this integrator.
        Graph.sparse_pwc_operator : Create `SparsePwc` operators.
        Graph.time_evolution_operators_pwc :
            Unitary time evolution operators for quantum systems with `Pwc` Hamiltonians.

        Notes
        -----
        The Lanczos algorithm calculates the unitary evolution of a state in the Krylov
        subspace. This subspace is spanned by the states resulting from applying the first
        `k` powers of the Hamiltonian on the input state, with `k` being the subspace dimension,
        much smaller that the full Hilbert space dimension. This allows for an efficient
        state propagation in high-dimensional systems compared to calculating the full
        unitary operator.

        Moreover, this function uses sparse matrix multiplication when the Hamiltonian is passed as
        a `SparsePwc`. This can lead to more efficient calculations when they involve large
        operators that are relatively sparse (contain mostly zeros). In this case, the initial state
        is still a densely represented array or tensor.

        Note that increasing the density of `sample_times` does not affect the accuracy of the
        integration. However, increasing the Krylov subspace dimension or subdividing the
        Hamiltonian into shorter piecewise-constant segments can reduce the integration error,
        at the expense of longer computation times.

        Examples
        --------
        See example in the `How to optimize controls on large sparse Hamiltonians
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-on-
        large-sparse-hamiltonians>`_ user guide.
        """

        dimension = initial_state.shape[-1]
        Checker.VALUE(
            hamiltonian.value_shape == (dimension, dimension),
            "The initial state and the Hamiltonian must have compatible shapes.",
            {
                "hamiltonian dimension": hamiltonian.value_shape,
                "initial_state.shape": initial_state.shape,
            },
        )

        if sample_times is None:
            shape = initial_state.shape
        else:
            duration = np.sum(hamiltonian.durations)
            sample_times = bounded_by(
                sample_times,
                "sample_times",
                duration,
                "the duration of Hamiltonian",
            )
            shape = (len(sample_times),) + initial_state.shape

        operation = create_operation(self.state_evolution_pwc, locals())
        return Tensor(operation, shape=shape)

    @validated
    def estimated_krylov_subspace_dimension_lanczos(
        self,
        spectral_range: Annotated[Union[float, Tensor], pipe(_scalar_float)],
        duration: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        maximum_segment_duration: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        error_tolerance: Annotated[float, pipe(ScalarT.REAL().gt(0))] = 1e-6,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate an appropriate Krylov subspace dimension (:math:`k`) to use in the Lanczos
        integrator while keeping the total error in the evolution below a given error tolerance.

        Note that you can provide your own estimation of the Hamiltonian spectral range or use the
        `spectral_range` operation to perform that calculation.

        Parameters
        ----------
        spectral_range : float or Tensor
            Estimated order of magnitude of Hamiltonian spectral range (difference
            between largest and smallest eigenvalues).
        duration : float
            The total evolution time.
        maximum_segment_duration : float
            The maximum duration of the piecewise-constant Hamiltonian segments.
        error_tolerance : float, optional
            Tolerance for the error in the integration, defined as the Frobenius norm of
            the vectorial difference between the exact state and the estimated state.
            Defaults to 1e-6.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            Recommended value of :math:`k` to use in a Lanczos integration with a Hamiltonian with a
            similar spectral range to the one passed.

        See Also
        --------
        Graph.spectral_range : Range of the eigenvalues of a Hermitian operator.
        Graph.state_evolution_pwc : Evolve a quantum state.

        Notes
        -----
        To provide the recommended :math:`k` parameter, this function uses the bound
        in the error for the Lanczos algorithm [1]_ [2]_ as an estimate for the error.
        For a single time step this gives

        .. math::
            \mathrm{error} \leq 12 \exp \left( - \frac{(w\tau)^2}{16 k} \right)
            \left (\frac{e w \tau}{ 4 k}  \right )^{k}

        where :math:`\tau` is the time step and :math:`w` is the spectral range of the Hamiltonian.

        As this bound overestimates the error, the actual resulting errors
        with the recommended parameter are expected to be (a few orders of magnitude)
        smaller than the requested tolerance.

        References
        ----------
        .. [1] `N. Del Buono and L. Lopez,
            Lect. Notes Comput. Sci. 2658, 111 (2003).
            <https://doi.org/10.1007/3-540-44862-4_13>`_

        .. [2] `M. Hochbruck and C. Lubich,
            SIAM J. Numer. Anal. 34, 1911 (1997).
            <https://doi.org/10.1137/S0036142995280572>`_

        Examples
        --------
        >>> graph.estimated_krylov_subspace_dimension_lanczos(
        ...     spectral_range=30.0,
        ...     duration=5e-7,
        ...     maximum_segment_duration=2.5e-8,
        ...     error_tolerance=1e-5,
        ...     name="dim",
        ... )
        <Tensor: name="dim", operation_name="estimated_krylov_subspace_dimension_lanczos", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="dim")
        >>> result["output"]["dim"]["value"]
        2

        See more examples in the `How to optimize controls on large sparse Hamiltonians
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-on-
        large-sparse-hamiltonians>`_ user guide.
        """

        Checker.VALUE(
            maximum_segment_duration <= duration,
            "The maximum segment duration must be less than or equal to duration.",
            {
                "maximum_segment_duration": maximum_segment_duration,
                "duration": duration,
            },
        )
        operation = create_operation(self.estimated_krylov_subspace_dimension_lanczos, locals())
        return Tensor(operation, shape=())

    @validated
    def spectral_range(
        self,
        operator: Annotated[
            Union[Tensor, np.ndarray, spmatrix],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
        iteration_count: Annotated[int, pipe(ScalarT.INT().gt(0))] = 3000,
        seed: Optional[Annotated[int, pipe(ScalarT.INT())]] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Obtain the range of the eigenvalues of a Hermitian operator.

        This function provides an estimate of the difference between the
        highest and the lowest eigenvalues of the operator. You can adjust its
        precision by modifying its default parameters.

        Parameters
        ----------
        operator : np.ndarray or scipy.sparse.spmatrix or Tensor
            The Hermitian operator :math:`M` whose range of eigenvalues you
            want to determine.
        iteration_count : int, optional
            The number of iterations :math:`N` in the calculation. Defaults to
            3000. Choose a higher number to improve the precision, or a smaller
            number to make the estimation run faster.
        seed : int or None, optional
            The random seed that the function uses to choose the initial random
            vector :math:`\left| r \right\rangle`. Defaults to None, which
            means that the function uses a different seed in each run.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor (scalar, real)
            The difference between the largest and the smallest eigenvalues of
            the operator.

        Warnings
        --------
        This calculation can be expensive, so we recommend that you run it
        before the optimization, if possible. You can do this by using a
        representative or a worst-case operator.

        Notes
        -----
        This function repeatedly multiplies the operator :math:`M` with a
        random vector :math:`\left| r \right\rangle`. In terms of the operator's
        eigenvalues :math:`\{ v_i \}` and eigenvectors
        :math:`\{\left|v_i \right\rangle\}`, the result of :math:`N` matrix
        multiplications is:

        .. math::
            M^N \left|r\right\rangle = \sum_i v_i^N \left|v_i\right\rangle
            \left\langle v_i \right. \left| r \right\rangle.

        For large :math:`N`, the term corresponding to the eigenvalue with
        largest absolute value :math:`V` will dominate the sum, as long as
        :math:`\left|r\right\rangle` has a non-zero overlap with its
        eigenvector. The function then retrieves the eigenvalue :math:`V` via:

        .. math::
            V \approx \frac{\left\langle r \right| M^{2N+1} \left| r
            \right\rangle}{\left\| M^N \left| r \right\rangle \right\|^2}.

        The same procedure applied to the matrix :math:`M-V` allows the function
        to find the eigenvalue at the opposite end of the spectral range.

        Examples
        --------
        >>> operator = np.diag([10, 40])
        >>> graph.spectral_range(operator, name="spectral_range")
        <Tensor: name="spectral_range", operation_name="spectral_range", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="spectral_range")
        >>> result["output"]["spectral_range"]["value"]
        30.0

        See more examples in the `How to optimize controls on large sparse Hamiltonians
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-on
        -large-sparse-hamiltonians>`_ user guide.
        """
        is_hermitian = True
        if issparse(operator):
            is_hermitian = np.allclose((operator - operator.getH()).data, 0)  # type: ignore
        elif isinstance(operator, np.ndarray):
            is_hermitian = np.allclose(operator, operator.T.conj())
        Checker.VALUE(is_hermitian, "The Hamiltonian must be Hermitian.")

        operation = create_operation(self.spectral_range, locals())
        return Tensor(operation, shape=())



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import (
    Callable,
    Optional,
    Union,
    overload,
)

import numpy as np
from qctrlcommons.node.wrapper import Operation

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.node_data import (
    Pwc,
    Stf,
    Tensor,
)
from boulderopal._nodes.utils import get_arraylike_shape
from boulderopal._nodes.validation import (
    Checker,
    ShapeT,
    number_or_shapeable,
    real_array,
    real_shapeable,
    shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    pipe,
    validated,
)


def _create_flexible_unary_node_data(
    _operation: Operation,
    x: int | float | complex | np.ndarray | Tensor | Pwc | Stf,
    name: Optional[str],
    value_shape_changer: Optional[Callable] = None,
) -> Tensor | Pwc | Stf:
    """
    Create the corresponding dataclass for nodes acting on Tensors, Pwcs, and Stfs
    implementing unary functions.

    Parameters
    ----------
    _operation : Operation
        The operation to implement.
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object on which the operation acts.
    name : str or None
        The name of the node.
    value_shape_changer : Callable[[tuple], tuple], optional
        Callable that transforms the original shape of the object into the shape after the operation
        is applied. Defaults to an identity operation, that is to say, to not change the shape.

    Returns
    -------
    Tensor or Pwc or Stf
        The operation acting on the object.
    """

    # By default don't change shapes.
    def get_value_shape(shape: tuple[int, ...]) -> tuple[int, ...]:
        if value_shape_changer is None:
            return shape
        return value_shape_changer(shape)

    if isinstance(x, Stf):
        Checker.VALUE(name is None, "You can't assign a name to an Stf node.", {"name": name})
        return Stf(
            _operation,
            value_shape=get_value_shape(x.value_shape),
            batch_shape=x.batch_shape,
        )

    if isinstance(x, Pwc):
        return Pwc(
            _operation,
            value_shape=get_value_shape(x.value_shape),
            durations=x.durations,
            batch_shape=x.batch_shape,
        )
    shape = get_arraylike_shape(x)
    return Tensor(_operation, shape=get_value_shape(shape))


class ArithmeticUnaryGraph:
    """
    Base class implementing unary arithmetic graph methods.
    """

    @overload
    def sqrt(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def sqrt(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def sqrt(self, x: Stf) -> Stf: ...

    @validated
    def sqrt(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise square root of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose square root you want to calculate, :math:`x`. For numbers, arrays, and
            tensors, the object is converted to a tensor and then the operation is applied. For
            functions of time (Pwcs and Stfs), the composition of the operation with the function
            is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise square root, :math:`\sqrt{x}`, of the values or function you provided.
            The returned object is of the same kind as the one you provided, except if you provide a
            number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.sqrt, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def sin(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def sin(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def sin(self, x: Stf) -> Stf: ...

    @validated
    def sin(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise sine of an object. This can be a number, an array, a tensor,
        or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose sine you want to calculate, :math:`x`. For numbers, arrays, and
            tensors, the object is converted to a tensor and then the operation is applied. For
            functions of time (Pwcs and Stfs), the composition of the operation with the function
            is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise sine, :math:`\sin{x}`, of the values or function you provided.
            The returned object is of the same kind as the one you provided, except if you provide a
            number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.sin, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def cos(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def cos(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def cos(self, x: Stf) -> Stf: ...

    @validated
    def cos(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise cosine of an object. This can be a number, an array, a tensor,
        or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose cosine you want to calculate, :math:`x`. For numbers, arrays, and
            tensors, the object is converted to a tensor and then the operation is applied. For
            functions of time (Pwcs and Stfs), the composition of the operation with the function
            is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise cosine, :math:`\cos{x}`, of the values or function you provided.
            The returned object is of the same kind as the one you provided, except if you provide a
            number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.cos, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def tan(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def tan(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def tan(self, x: Stf) -> Stf: ...

    @validated
    def tan(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise tangent of an object. This can be a number, an array, a tensor,
        or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose tangent you want to calculate, :math:`x`. For numbers, arrays, and
            tensors, the object is converted to a tensor and then the operation is applied. For
            functions of time (Pwcs and Stfs), the composition of the operation with the function
            is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise tangent, :math:`\tan{x}`, of the values or function you provided.
            The returned object is of the same kind as the one you provided, except if you provide a
            number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.tan, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def sinh(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def sinh(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def sinh(self, x: Stf) -> Stf: ...

    @validated
    def sinh(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise hyperbolic sine of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose hyperbolic sine you want to calculate, :math:`x`.
            For numbers, arrays, and tensors, the object is converted to a tensor and
            then the operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise hyperbolic sine, :math:`\sinh{x}`, of the values or function you
            provided.
            The returned object is of the same kind as the one you provided, except if you provide a
            number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.sinh, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def cosh(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def cosh(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def cosh(self, x: Stf) -> Stf: ...

    @validated
    def cosh(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise hyperbolic cosine of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose hyperbolic cosine you want to calculate, :math:`x`.
            For numbers, arrays, and tensors, the object is converted to a tensor and
            then the operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise hyperbolic cosine, :math:`\cosh{x}`, of the values or function you
            provided. The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.cosh, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def tanh(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def tanh(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def tanh(self, x: Stf) -> Stf: ...

    @validated
    def tanh(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise hyperbolic tangent of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose hyperbolic tangent you want to calculate, :math:`x`.
            For numbers, arrays, and tensors, the object is converted to a tensor and
            then the operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise hyperbolic tangent, :math:`\tanh{x}`, of the values or function you
            provided. The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.tanh, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def log(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def log(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def log(self, x: Stf) -> Stf: ...

    @validated
    def log(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise natural logarithm of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose natural logarithm you want to calculate, :math:`x`.
            For numbers, arrays, and tensors, the object is converted to a tensor and then the
            operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise natural logarithm, :math:`\log{x}`, of the values or function you
            provided. The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.log, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def exp(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def exp(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def exp(self, x: Stf) -> Stf: ...

    @validated
    def exp(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise exponential of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose exponential you want to calculate, :math:`x`. For numbers, arrays,
            and tensors, the object is converted to a tensor and then the operation is applied. For
            functions of time (Pwcs and Stfs), the composition of the operation with the function
            is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise exponential, :math:`e^{x}`, of the values or function you
            provided. The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.exp, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def negative(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def negative(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def negative(self, x: Stf) -> Stf: ...

    @validated
    def negative(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise numerical negative value of an object. This can be a number,
        an array, a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose numerical negative value you want to calculate, :math:`x`. For numbers,
            arrays, and tensors, the object is converted to a tensor and then the operation is
            applied. For functions of time (Pwcs and Stfs), the composition of the operation with
            the function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise negation, :math:`-x`, of the values or function you
            provided. The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.negative, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def real(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def real(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def real(self, x: Stf) -> Stf: ...

    @validated
    def real(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise real part of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose real part you want to calculate, :math:`x`. For numbers,
            arrays, and tensors, the object is converted to a tensor and then the operation is
            applied. For functions of time (Pwcs and Stfs), the composition of the operation with
            the function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise real part, :math:`\Re(x)`, of the values or function you provided. The
            returned object is a real object of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.

        See Also
        --------
        Graph.complex_value : Create a complex object from its real and imaginary parts.
        Graph.imag : Imaginary part of a complex object.
        """
        operation = create_operation(self.real, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def imag(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def imag(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def imag(self, x: Stf) -> Stf: ...

    @validated
    def imag(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise imaginary part of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose imaginary part you want to calculate, :math:`x`. For numbers,
            arrays, and tensors, the object is converted to a tensor and then the operation is
            applied. For functions of time (Pwcs and Stfs), the composition of the operation with
            the function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise imaginary part, :math:`\Im(x)`, of the values or function you provided.
            The returned object is a real object of the same kind as the one you provided, except if
            you provide a number or an np.ndarray in which case it's a Tensor.

        See Also
        --------
        Graph.complex_value : Create a complex object from its real and imaginary parts.
        Graph.real : Real part of a complex object.
        """
        operation = create_operation(self.imag, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def abs(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def abs(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def abs(self, x: Stf) -> Stf: ...

    @validated
    def abs(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise absolute value of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose absolute value you want to calculate, :math:`x`. For numbers,
            arrays, and tensors, the object is converted to a tensor and then the operation is
            applied. For functions of time (Pwcs and Stfs), the composition of the operation with
            the function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise absolute value, :math:`\left|x\right|`, of the values or function you
            provided. The returned object is a real object of the same kind as the one you provided,
            except if you provide a number or an np.ndarray in which case it's a Tensor.

        See Also
        --------
        Graph.angle : Argument of a complex object.
        Graph.complex_value : Create a complex object from its real and imaginary parts.
        """
        operation = create_operation(self.abs, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def angle(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def angle(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def angle(self, x: Stf) -> Stf: ...

    @validated
    def angle(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise argument of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose argument you want to calculate, :math:`x`. For numbers,
            arrays, and tensors, the object is converted to a tensor and then the operation is
            applied. For functions of time (Pwcs and Stfs), the composition of the operation with
            the function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise argument, :math:`\arg(x)`, of the values or function you provided. The
            returned object is a real object of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.

        See Also
        --------
        Graph.abs : Absolute value of a complex object.
        Graph.complex_value : Create a complex object from its real and imaginary parts.
        """
        operation = create_operation(self.angle, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def conjugate(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def conjugate(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def conjugate(self, x: Stf) -> Stf: ...

    @validated
    def conjugate(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise complex conjugate of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The object whose complex conjugate you want to calculate, :math:`x`. For numbers,
            arrays, and tensors, the object is converted to a tensor and then the operation is
            applied. For functions of time (Pwcs and Stfs), the composition of the operation with
            the function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise complex conjugate, :math:`x^\ast`, of the values or function you
            provided. The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.

        See Also
        --------
        Graph.adjoint : Hermitian adjoint of an operator.
        """
        operation = create_operation(self.conjugate, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def arcsin(
        self,
        x: Union[float, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def arcsin(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def arcsin(self, x: Stf) -> Stf: ...

    @validated
    def arcsin(
        self,
        x: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise arcsine of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : float or np.ndarray or Tensor or Pwc or Stf
            The object whose arcsine you want to calculate, :math:`x`. Must be real.
            For numbers, arrays, and tensors, the object is converted to a tensor and then
            the operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise arcsine, :math:`\arcsin{x}`, of the values or function you
            provided. Outputs will be in the range of :math:`[-\pi/2, \pi/2]`.
            The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.arcsin, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def arccos(
        self,
        x: Union[float, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def arccos(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def arccos(self, x: Stf) -> Stf: ...

    @validated
    def arccos(
        self,
        x: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise arccosine of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : float or np.ndarray or Tensor or Pwc or Stf
            The object whose arccosine you want to calculate, :math:`x`. Must be real.
            For numbers, arrays, and tensors, the object is converted to a tensor and then
            the operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise arccosine, :math:`\arccos{x}`, of the values or function you
            provided. Outputs will be in the range of :math:`[0, \pi]`.
            The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.arccos, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def arctan(
        self,
        x: Union[float, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def arctan(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def arctan(self, x: Stf) -> Stf: ...

    @validated
    def arctan(
        self,
        x: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise arctangent of an object. This can be a number, an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf.

        Parameters
        ----------
        x : float or np.ndarray or Tensor or Pwc or Stf
            The object whose arctangent you want to calculate, :math:`x`. Must be real.
            For numbers, arrays, and tensors, the object is converted to a tensor and then
            the operation is applied.
            For functions of time (Pwcs and Stfs), the composition of the operation with the
            function is computed (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise arctangent, :math:`\arctan{x}`, of the values or function you
            provided. Outputs will be in the range of :math:`[-\pi/2, \pi/2]`.
            The returned object is of the same kind as the one you provided, except if you
            provide a number or an np.ndarray in which case it's a Tensor.
        """
        operation = create_operation(self.arctan, locals())
        return _create_flexible_unary_node_data(operation, x, name)

    @overload
    def adjoint(self, x: Union[np.ndarray, Tensor], *, name: Optional[str] = None) -> Tensor: ...
    @overload
    def adjoint(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def adjoint(self, x: Stf) -> Stf: ...

    @validated
    def adjoint(
        self,
        x: Annotated[Union[np.ndarray, Tensor, Pwc, Stf], pipe(shapeable, after=ShapeT.MATRIX())],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise adjoint of the last two dimensions of an object.
        This can be a an array, a tensor, or a time-dependent function in the form of a
        Pwc or an Stf where values have at least two dimensions.

        Parameters
        ----------
        x : np.ndarray or Tensor or Pwc or Stf
            The object whose adjoint you want to calculate, :math:`X^\dagger`.
            Must be a matrix or a matrix-valued function.
            For arrays and tensors, the object is converted to a tensor and then
            the operation is applied. For functions of time (Pwcs and Stfs), the composition
            of the operation with the function is computed
            (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise adjoint, of the last two dimension of the given matrix or matrix-valued
            function.

        See Also
        --------
        Graph.conjugate : Conjugate of a complex object.
        Graph.transpose : Reorder the dimensions of a tensor.
        """

        def value_shape_changer(value_shape: tuple[int, ...]) -> tuple[int, ...]:
            *batch, x, y = value_shape  # Unpacking syntax is awesome.
            return (*batch, y, x)

        operation = create_operation(self.adjoint, locals())
        return _create_flexible_unary_node_data(operation, x, name, value_shape_changer)

    @overload
    def trace(self, x: Union[np.ndarray, Tensor], *, name: Optional[str] = None) -> Tensor: ...
    @overload
    def trace(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def trace(self, x: Stf) -> Stf: ...

    @validated
    def trace(
        self,
        x: Annotated[Union[np.ndarray, Tensor, Pwc, Stf], pipe(shapeable, after=ShapeT.MATRIX())],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the trace of an object. This can be a an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf
        where values have at least two dimensions.
        The trace is calculated on the last two dimensions.

        Parameters
        ----------
        x : np.ndarray or Tensor or Pwc or Stf
            The object whose trace you want to calculate, :math:`\mathop{\mathrm{Tr}}(x)`.
            Must be a matrix or a matrix-valued function.
            For arrays and tensors, the object is converted to a tensor and then
            the operation is applied. For functions of time (Pwcs and Stfs), the composition
            of the operation with the function is computed
            (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The trace of the last two dimension of the given matrix or matrix-valued
            function. Outputs will have two fewer dimensions.

        See Also
        --------
        Graph.einsum : Tensor contraction via Einstein summation convention.
        """
        operation = create_operation(self.trace, locals())
        return _create_flexible_unary_node_data(
            operation,
            x,
            name,
            lambda value_shape: value_shape[:-2],
        )

    @overload
    def hermitian_part(
        self,
        x: Union[np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def hermitian_part(self, x: Pwc, *, name: Optional[str] = None) -> Pwc: ...
    @overload
    def hermitian_part(self, x: Stf) -> Stf: ...

    @validated
    def hermitian_part(
        self,
        x: Annotated[
            Union[np.ndarray, Tensor, Pwc, Stf],
            pipe(shapeable, after=ShapeT.OPERATOR()),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the Hermitian part of an object. This can be an array,
        a tensor, or a time-dependent function in the form of a Pwc or an Stf
        where values have at least two dimensions.
        The operation is applied on the last two dimensions, which must be equal to each other.

        Parameters
        ----------
        x : np.ndarray or Tensor or Pwc or Stf
            The object whose Hermitian part you want to calculate, :math:`\mathop{x}`.
            Must be a matrix or a matrix-valued function.
            For arrays and tensors, the object is converted to a tensor and then
            the operation is applied. For functions of time (Pwcs and Stfs), the composition
            of the operation with the function is computed
            (that is, the operation is applied to the function values).
        name : str or None, optional
            The name of the node. You can only provide a name if the object is not an Stf.

        Returns
        -------
        Tensor or Pwc or Stf
            The Hermitian part of the matrix or matrix-valued function,
            :math:`\frac{1}{2}(\mathop{x}+\mathop{x}^\dagger)`.
            Outputs will have the same dimension as the inputs.

        See Also
        --------
        Graph.adjoint : Hermitian adjoint of an operator.

        Examples
        --------
        Create a Hamiltonian from a non-Hermitian Pwc operator.

        >>> omega = graph.pwc(durations=np.array([0.5, 0.7]), values=np.array([0.2, 0.4]))
        >>> sigma_m = np.array([[0, 1], [0, 0]])
        >>> operator = omega * sigma_m
        >>> graph.hermitian_part(operator, name="hamiltonian")
        <Pwc: name="hamiltonian", operation_name="hermitian_part", value_shape=(2, 2), batch_shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
        >>> result["output"]["hamiltonian"]
        {
            'durations': array([0.5, 0.7]),
            'values': array([
                [[0. , 0.1], [0.1, 0. ]],
                [[0. , 0.2], [0.2, 0. ]]
                ]),
            'time_dimension': 0
        }

        See more examples in the `Simulate the dynamics of a single qubit using computational graphs
        <https://docs.q-ctrl.com/boulder-opal/tutorials/simulate-the-dynamics-of-a-single-qubit-
        using-computational-graphs>`_ tutorial.

        Create a Hamiltonian from a non-Hermitian Stf operator.

        >>> operator = stf_signal * np.array([[0, 1, 0], [0, 0, np.sqrt(2)], [0, 0, 0]])
        >>> hamiltonian = graph.hermitian_part(operator)
        >>> hamiltonian
        <Stf: operation_name="hermitian_part", value_shape=(3, 3), batch_shape=()>

        Create a Hermitian matrix from a non-Hermitian np.ndarray.

        >>> sigma_m = np.array([[0, 1], [0, 0]])
        >>> graph.hermitian_part(sigma_m, name="hamiltonian")
        <Tensor: name="hamiltonian", operation_name="hermitian_part", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
        >>> result["output"]["hamiltonian"]
        {'value': array([[0. , 0.5], [0.5, 0. ]])}
        """  # noqa: E501
        operation = create_operation(self.hermitian_part, locals())
        return _create_flexible_unary_node_data(operation, x, name)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    Optional,
    SupportsIndex,
    no_type_check,
)

import numpy as np
from qctrlcommons.node.wrapper import (
    NamedNodeData,
    NodeData,
    Operation,
)
from typing_extensions import Self

if TYPE_CHECKING:
    from boulderopal.graph import Graph


def _get_item(
    graph: Graph,
    value: Any,
    key: SupportsIndex | slice,
    name: Optional[str] = None,
) -> Any:
    """
    Get an item (or items) from a node value.

    Typically you would use slicing syntax ``value[key]`` instead of using this function directly.

    Parameters
    ----------
    graph : Graph
        The graph that the node belongs to.
    value : Any
        The value from which to get the item.
    key : SupportsIndex or slice
        The key for the item or items.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Any
        The item or items obtained from ``value[key]``.
    """

    return NamedNodeData(Operation(graph, "getitem", False, value=value, key=key, name=name))


def _get_attribute(graph: Graph, value: Any, attr: str, name: Optional[str] = None) -> Any:
    """
    Get an attribute from a node value.

    Typically you would use the syntax ``value.attr`` instead of using this function directly.

    Parameters
    ----------
    graph : Graph
        The graph that the node belongs to.
    value : Any
        The value from which to get the item.
    attr : str
        The name of the attribute.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Any
        The item or items obtained from ``value.attr``, or to be more precise,
        ``getattr(value, attr)``.

    Notes
    -----
    Only certain combinations of `value` and `attr` are supported.
    """

    valid = {"FilterFunction": ("inverse_powers", "uncertainties"), "Pwc": ("values")}
    if attr in valid.get(value.__class__.__name__, ()):
        return NamedNodeData(Operation(graph, "getattr", False, value=value, attr=attr, name=name))

    # Invalid combination of value type and attr, so show an error message.
    raise ValueError(
        f"Not allowed to fetch attribute. Valid combinations are {valid}, "
        f"but got {value.__class__.__name__} with {attr}.",
    )


class ArithmeticMixin:
    """
    Mixin to be used by NodeData that support binary arithmetic operations with
    number/array/Tensor/Pwc/Stf objects.

    By default ``arr + graph.op()`` throws an error, since NumPy doesn't know how to add
    `graph.op()` objects to arrays. Even the fact that the func ops override `__radd__` doesn't
    help, since the NumPy addition takes precedence. We can instead tell NumPy to delegate all
    binary operations to the other operand, by explicitly clearing the `__array_ufunc__`
    attribute.
    """

    __array_ufunc__ = None

    @no_type_check
    def __add__(self, other) -> Self:
        return self.operation.graph.add(self, other)

    @no_type_check
    def __radd__(self, other) -> Self:
        return self.operation.graph.add(other, self)

    @no_type_check
    def __sub__(self, other) -> Self:
        return self.operation.graph.subtract(self, other)

    @no_type_check
    def __rsub__(self, other) -> Self:
        return self.operation.graph.subtract(other, self)

    @no_type_check
    def __matmul__(self, other) -> Self:
        return self.operation.graph.matmul(self, other)

    @no_type_check
    def __rmatmul__(self, other) -> Self:
        return self.operation.graph.matmul(other, self)

    @no_type_check
    def __mul__(self, other) -> Self:
        return self.operation.graph.multiply(self, other)

    @no_type_check
    def __rmul__(self, other) -> Self:
        return self.operation.graph.multiply(other, self)

    @no_type_check
    def __floordiv__(self, other) -> Self:
        return self.operation.graph.floordiv(self, other)

    @no_type_check
    def __rfloordiv__(self, other) -> Self:
        return self.operation.graph.floordiv(other, self)

    @no_type_check
    def __pow__(self, power) -> Self:
        return self.operation.graph.pow(self, power)

    @no_type_check
    def __rpow__(self, other) -> Self:
        return self.operation.graph.pow(other, self)

    @no_type_check
    def __truediv__(self, other) -> Self:
        return self.operation.graph.truediv(self, other)

    @no_type_check
    def __rtruediv__(self, other) -> Self:
        return self.operation.graph.truediv(other, self)

    @no_type_check
    def __abs__(self) -> Self:
        return self.operation.graph.abs(self)

    @no_type_check
    def __neg__(self) -> Self:
        return self.operation.graph.negative(self)


class Tensor(NamedNodeData, ArithmeticMixin):
    """
    A multi-dimensional array of data.

    Most functions accepting a :obj:`.Tensor` object can alternatively accept a NumPy array.

    You can use the arithmetic operators ``+``, ``-``, ``*``, ``**``, ``/``, ``//``, and ``@``
    to perform operations between two `Tensor` objects.

    Attributes
    ----------
    shape : tuple
        The shape of the tensor.
    name : str
        The name assigned to the node.

    See Also
    --------
    Graph.tensor : Create a real or complex Tensor with the data provided.

    Notes
    -----
    The value of a `Tensor` node can be fetched in a graph calculation
    by adding its `name` in the `output_node_names` parameter for the function call.
    This will add an item to the output dictionary in the calculation result object,
    whose key is `name`. The item's value will be a dictionary
    with the "value" of the Tensor as a NumPy array.
    """

    def __init__(self, operation: Operation, shape: tuple[int, ...]) -> None:
        self.shape = shape
        super().__init__(operation=operation)
        self.operation.is_scalar_tensor = np.prod(self.shape) == 1

    def __getitem__(self, item: SupportsIndex | slice) -> Tensor:
        """
        Refer to item in operation.
        """
        node_data = _get_item(self.operation.graph, self, item)
        shape = np.empty(self.shape)[item].shape
        return Tensor(node_data.operation, shape=shape)

    def __iter__(self) -> None:
        # Disable iteration for now. Even though this should work fine in theory (since all client
        # tensors have fully-defined shapes), allowing iterability on the client causes tensors to
        # pass checks that will fail in the backend (for example, if tensors are iterable on the
        # client, a multi-dimensional tensor can be passed to a function that expects a list of
        # tensors; such an input will fail in the backend though). This could be revisited in the
        # future if we're more strict about client-side validation of iterable inputs, or if we
        # update the backend to be able to iterate over tensors.
        raise TypeError(
            "You cannot iterate over Tensors directly. Instead you can iterate over the indices "
            "and extract elements of the tensor using `tensor[index]`.",
        )

    def __repr__(self) -> str:
        return (
            f'<Tensor: name="{self.name}", '
            f'operation_name="{self.operation.operation_name}", '
            f"shape={self.shape}>"
        )


class Pwc(NamedNodeData, ArithmeticMixin):
    """
    A piecewise-constant tensor-valued function of time (or batch of such functions).

    You can use the arithmetic operators ``+``, ``-``, ``*``, ``**``, ``/``, ``//``, and ``@``
    to perform operations between two `Pwc` objects or between a `Pwc` and a `Tensor`.

    Attributes
    ----------
    values : Tensor
        The values of the function on the piecewise-constant segments.
    durations : np.ndarray
        The durations of the constant segments.
    value_shape : tuple
        The shape of the function value.
    batch_shape : tuple
        The shape of the batch in the function.
    name : str
        The name assigned to the node.

    See Also
    --------
    Graph.pwc : Operation to create piecewise-constant functions.

    Notes
    -----
    The value of a `Pwc` node can be fetched in a graph calculation
    by adding its `name` in the `output_node_names` parameter for the function call.
    This will add an item to the output dictionary in the calculation result object,
    whose key is `name`. The item's value will be a dictionary with the "durations"
    of the piecewise-constant segments, the "values" of the function at each segment,
    and its "time_dimension" (the axis of the values array that corresponds to time).
    Dimensions of the values trailing the time dimension represent the dimensions of
    the object represented by the `Pwc`.
    Dimensions preceding the time dimension represent batch dimensions.

    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    def __init__(
        self,
        operation: Operation,
        value_shape: tuple[int, ...],
        durations: np.ndarray,
        batch_shape: tuple[int, ...],
    ) -> None:
        self.value_shape = value_shape
        self.batch_shape = batch_shape
        self.durations = durations
        super().__init__(operation=operation)

    @property
    def values(self) -> Tensor:
        """
        Access to the values in Pwc.
        """
        node_data = _get_attribute(self.operation.graph, self, "values")
        shape = tuple(self.batch_shape) + (len(self.durations),) + tuple(self.value_shape)
        return Tensor(node_data.operation, shape=shape)

    def __repr__(self) -> str:
        return (
            f'<Pwc: name="{self.name}", '
            f'operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}, batch_shape={self.batch_shape}>"
        )


class Stf(NodeData, ArithmeticMixin):
    """
    A sampleable tensor-valued function of time (or batch of such functions).

    You can use the arithmetic operators ``+``, ``-``, ``*``, ``**``, ``/``, ``//``, and ``@``
    to perform operations between two `Stf` objects or between an `Stf` and a `Tensor`.

    Attributes
    ----------
    value_shape : tuple
        The shape of the function value.
    batch_shape : tuple
        The shape of the batch in the function.

    See Also
    --------
    Graph.identity_stf : Operation to create an `Stf` representing the identity function.

    Notes
    -----
    Stf nodes represent arbitrary functions of time. Piecewise-constant (PWC) or constant functions
    are special cases of Stfs and the Q-CTRL Python package provides specific APIs to support them.
    Note that as the PWC property can simplify the calculation, you should always consider using
    PWC-related APIs if your system parameters or controls are described by PWC functions.

    The value of `Stf` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.

    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    def __init__(
        self,
        operation: Operation,
        value_shape: tuple[int, ...],
        batch_shape: tuple[int, ...],
    ):
        self.value_shape = value_shape
        self.batch_shape = batch_shape
        super().__init__(operation=operation)

    def __repr__(self) -> str:
        return (
            f'<Stf: operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}, batch_shape={self.batch_shape}>"
        )


class SparsePwc(NodeData):
    """
    A piecewise-constant sparse-matrix-valued function of time.

    Attributes
    ----------
    value_shape : tuple[int, ...]
        The shape of the function value.
    durations : np.ndarray
        The durations of the constant segments.

    See Also
    --------
    Graph.sparse_pwc_operator : Operation to create `SparsePwc` operators.

    Notes
    -----
    The value of `SparsePwc` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.
    """

    def __init__(
        self,
        operation: Operation,
        value_shape: tuple[int, ...],
        durations: np.ndarray,
    ) -> None:
        self.value_shape = value_shape
        self.durations = durations
        super().__init__(operation=operation)

    def __repr__(self) -> str:
        return (
            f'<SparsePwc: operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}>"
        )


class Target(NodeData):
    """
    A target gate for an infidelity calculation.

    Attributes
    ----------
    value_shape : tuple[int, ...]
        The shape of the target gate operation.

    See Also
    --------
    Graph.target : Operation to define the target operation of a time evolution.

    Notes
    -----
    The value of `Target` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.
    """

    def __init__(self, operation: Operation, value_shape: tuple[int, ...]) -> None:
        self.value_shape = value_shape
        super().__init__(operation=operation)

    def __repr__(self) -> str:
        return (
            f'<Target: operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}>"
        )


class ConvolutionKernel(NodeData):
    """
    A kernel to be used in a convolution.

    See Also
    --------
    Graph.convolve_pwc : Operation to create an `Stf` by convolving a `Pwc` with a kernel.
    Graph.gaussian_convolution_kernel : Operation to create a convolution kernel representing a
        normalized Gaussian.
    Graph.sinc_convolution_kernel : Operation to create a convolution kernel representing the sinc
        function.

    Notes
    -----
    The value of `ConvolutionKernel` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.
    """

    def __init__(self, operation: Operation) -> None:
        super().__init__(operation=operation)

    def __repr__(self) -> str:
        return f'<ConvolutionKernel: operation_name="{self.operation.operation_name}">'


class Sequence(NamedNodeData):
    """
    Wrapper class for creating a sequence of Nodes.
    """

    def __init__(self, operation: Operation) -> None:
        self.items: list[Any] = []
        super().__init__(operation=operation)

    def create_sequence(self, node_constructor: Callable, size: int) -> Sequence:
        """
        Populate the `items` of the sequence from the operation.

        Parameters
        ----------
        node_constructor : Callable
            A callable to generate the node data for the sequence.
        size : int
            Size of the sequence.

        Returns
        -------
        Sequence
            The sequence itself.
        """

        def get_item_op(item: SupportsIndex) -> Operation:
            return _get_item(self.operation.graph, self, item).operation

        self.items = [node_constructor(get_item_op(index), index) for index in range(size)]

        return self

    def __getitem__(self, index: int) -> Operation:
        return self.items[index]

    def __iter__(self) -> Iterator:
        return iter(self.items)

    def __len__(self) -> int:
        return len(self.items)

    def __repr__(self) -> str:
        return f'<Sequence: operation_name="{self.operation.operation_name}" f"items={self.items}>'


class FilterFunction(NamedNodeData):
    """
    A tensor-valued filter function result.

    Attributes
    ----------
    inverse_powers : Tensor
        The values of the filter function at the given frequencies.
    uncertainties : Tensor, optional
        The uncertainties of the filter function values.
        This field is None when the exact method is used for computing the filter function.
    frequencies : np.ndarray
        The frequencies at which the filter function has been evaluated.
    value_shape : tuple[int]
        The shape of the function value.
    exact : bool
        Indicates whether filter function is exact.
    name : str
        The name assigned to the node.

    See Also
    --------
    Graph.filter_function :
        Evaluate the filter function for a control Hamiltonian and a noise operator at the given
        frequency elements.

    Notes
    -----
    The value of a `FilterFunction` node can be fetched in a graph calculation
    by adding its `name` in the `output_node_names` parameter for the function call.
    This will add an item to the output dictionary in the calculation result object,
    whose key is `name`. The item's value will be a dictionary
    with NumPy arrays with the "frequencies" and "inverse_powers" values.
    If the filter function calculation is not exact, the dictionary also
    contains an array with the filter function "uncertainties".
    """

    def __init__(self, operation: Operation, frequencies: np.ndarray, exact: bool) -> None:
        self.frequencies = frequencies
        self.exact = exact
        super().__init__(operation=operation)

    @property
    def value_shape(self) -> tuple[int, ...]:
        """
        Access the value shape.
        """
        return (len(self.frequencies),)

    @property
    def inverse_powers(self) -> Tensor:
        """
        Access to the inverse powers in FilterFunction.
        """
        node_data = _get_attribute(self.operation.graph, self, "inverse_powers")
        return Tensor(node_data.operation, shape=self.value_shape)

    @property
    def uncertainties(self) -> Optional[Tensor]:
        """
        Access to the uncertainties in FilterFunction.
        """
        if self.exact is False:
            node_data = _get_attribute(self.operation.graph, self, "uncertainties")
            return Tensor(node_data.operation, shape=self.value_shape)
        return None

    def __repr__(self) -> str:
        return (
            f'<FilterFunction: name="{self.name}", '
            f'operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}>"
        )



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from enum import Enum
from functools import partial
from typing import (
    Any,
    Optional,
    Type,
)

import numpy as np


def _simple_check(
    condition: bool | np.bool_,
    message: str,
    extras: Optional[dict[str, Any]] = None,
    *,
    exc: Type[Exception],
) -> None:
    if condition:
        return
    if extras is not None:
        _extra = ", ".join([f"{key}={val!r}" for key, val in extras.items()])
        message = f"{message} {_extra}"
    raise exc(message)


class Checker(Enum):
    """
    Host validation functions to check input for Boulder Opal APIs.
    """

    TYPE = partial(_simple_check, exc=TypeError)
    VALUE = partial(_simple_check, exc=ValueError)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        return self.value(*args, **kwargs)
